#!/usr/bin/python
# vim:ts=4

import sys, os, string, re, pwd
import pg
from time import time, sleep
from popen2 import popen2
from LATClient import LATClient
from VendingMachine import VendingMachine
from ConfigParser import ConfigParser

GREETING = 'UCC SNACKS'
PIN_LENGTH = 4

DOOR = 1
SWITCH = 2
KEY = 3

class DispenseDatabase:
	def __init__(self, vending_machine, host, name, user, password):
		self.vending_machine = vending_machine
		self.db = pg.DB(dbname = name, host = host, user = user, passwd = password)
		self.db.query('LISTEN vend_requests')

	def process_requests(self):
		print 'processing'
		query = 'SELECT request_id, request_slot FROM vend_requests WHERE request_handled = false'
		try:
			outstanding = self.db.query(query).getresult()
		except (pg.error,), db_err:
			sys.stderr.write('Failed to query database: %s\n'%(db_err.strip()))
			return
		for (id, slot) in outstanding:
			(worked, code, string) = self.vending_machine.vend(slot)
			print (worked, code, string)
			if worked:
				query = 'SELECT vend_success(%s)'%id
				self.db.query(query).getresult()
			else:
				query = 'SELECT vend_failed(%s)'%id
				self.db.query(query).getresult()

	def handle_events(self):
		notifier = self.db.getnotify()
		while notifier is not None:
			self.process_requests()
			notify = self.db.getnotify()

def get_pin(uid):
	try:
		info = pwd.getpwuid(uid)
	except KeyError:
		return None
	if info.pw_dir == None: return False
	pinfile = os.path.join(info.pw_dir, '.pin')
	try:
		s = os.stat(pinfile)
	except OSError:
		return None
	if s.st_mode & 077:
		return None
	try:
		f = file(pinfile)
	except IOError:
		return None
	pinstr = f.readline()
	f.close()
	if not re.search('^'+'[0-9]'*PIN_LENGTH+'$', pinstr):
		return None
	return int(pinstr)

def has_good_pin(uid):
	return get_pin(uid) != None

def verify_user_pin(uid, pin):
	if get_pin(uid) == pin:
		info = pwd.getpwuid(uid)
		return info.pw_name
	else:
		return None

def door_open_mode(vending_machine):
	print "Entering open door mode"
	v.display("DOOR  OPEN")
	while True:
		e = v.next_event()
		if e == None: break
		(event, params) = e
		if event == DOOR:
			if params == 1: # door closed
				v.display("BYE BYE!")
				sleep(1)
				return

def center(str):
	LEN = 10
	return ' '*((LEN-len(str))/2)+str

class MessageKeeper:
	def __init__(self, vendie):
		self.scrolling_message = []
		self.v = vendie
		self.next_update = None

	def set_message(self, string):
		self.scrolling_message = [(string, False, None)]
		self.update_display(True)

	def set_messages(self, strings):
		self.scrolling_message = strings
		self.update_display(True)

	def update_display(self, forced = False):
		if not forced and self.next_update != None and time() < self.next_update:
			return
		if len(self.scrolling_message) > 0:
			newmsg = self.scrolling_message[0]
			if newmsg[2] != None:
				self.next_update = time() + newmsg[2]
			else:
				self.next_update = None
			self.v.display(self.scrolling_message[0][0])
			if self.scrolling_message[0][1]:
				self.scrolling_message.append(self.scrolling_message[0])
			del self.scrolling_message[0]

if __name__ == '__main__':
	cp = ConfigParser()
	cp.read('/etc/dispense/servers.conf')
	DBServer = cp.get('Database', 'Server')
	DBName = cp.get('Database', 'Name')
	DBUser = cp.get('VendingMachine', 'DBUser')
	DBPassword = cp.get('VendingMachine', 'DBPassword')

	ServiceName = cp.get('VendingMachine', 'ServiceName')
	ServicePassword = cp.get('VendingMachine', 'Password')
	# Open vending machine via LAT
	if 0:
		latclient = LATClient(service = ServiceName, password = ServicePassword)
		(rfh, wfh) = latclient.get_fh()
	else:
		#(rfh, wfh) = popen2('../../virtualvend/vvend.py')
		import socket
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
		sock.connect(('localhost', 5150))
		rfh = sock.makefile('r')
		wfh = sock.makefile('w')
	v = VendingMachine(rfh, wfh)
	print 'PING is', v.ping()

	db = DispenseDatabase(v, DBServer, DBName, DBUser, DBPassword)
	cur_user = ''
	cur_pin = ''
	cur_selection = ''

	mk = MessageKeeper(v)
	mk.set_message(GREETING)
	logout_timeout = None
	last_timeout_refresh = None

	while True:
		db.handle_events()

		if logout_timeout != None:
			time_left = logout_timeout - time()
			if time_left < 10 and last_timeout_refresh > time_left:
				mk.set_message('LOGOUT: '+str(int(time_left)))
				last_timeout_refresh = int(time_left)

		if logout_timeout != None and logout_timeout - time() <= 0:
			logout_timeout = None
			cur_user = ''
			cur_pin = ''
			cur_selection = ''
			mk.set_message(GREETING)

		mk.update_display()

		e = v.next_event(0)
		if e == None:
			e = v.next_event(0.1)
			if e == None:
				continue
		(event, params) = e
		print e
		if event == DOOR:
			if params == 0:
				door_open_mode(v);
				cur_user = ''
				cur_pin = ''
				mk.set_message(GREETING)
		elif event == SWITCH:
			# don't care right now.
			pass
		elif event == KEY:
			key = params
			# complicated key handling here:
			if len(cur_user) < 5:
				if key == 11:
					cur_user = ''
					mk.set_message(GREETING)
					continue
				cur_user += chr(key + ord('0'))
				mk.set_message('UID: '+cur_user)
				if len(cur_user) == 5:
					uid = int(cur_user)
					if not has_good_pin(uid):
						mk.set_messages(
							[(center('INVALID'), False, 0.7),
							 (center('PIN'), False, 0.7),
							 (center('SETUP'), False, 1.0),
							 (GREETING, False, None)])
						cur_user = ''
						cur_pin = ''
						continue
					cur_pin = ''
					mk.set_message('PIN: ')
					continue
			elif len(cur_pin) < PIN_LENGTH:
				if key == 11:
					if cur_pin == '':
						cur_user = ''
						mk.set_message(GREETING)
						continue
					cur_pin = ''
					mk.set_message('PIN: ')
					continue
				cur_pin += chr(key + ord('0'))
				mk.set_message('PIN: '+'X'*len(cur_pin))
				if len(cur_pin) == PIN_LENGTH:
					username = verify_user_pin(int(cur_user), int(cur_pin))
					if username:
						v.beep(0, False)
						cur_selection = ''

						msg = [(center('WELCOME'), False, 0.8),
							   (center(username), False, 0.9)]
						msg.append(('CHOICES :', True, 1))
						msg.append(('55 - DOOR', True, 1))
						msg.append(('OR A SNACK', True, 1))
						mk.set_messages(msg)
						continue
					else:
						v.beep(40, False)
						mk.set_messages(
							[(center('BAD PIN'), False, 1.0),
							 (center('SORRY'), False, 0.5),
							 (GREETING, False, None)])
						cur_user = ''
						cur_pin = ''
						continue
			elif len(cur_selection) == 0:
				if key == 11:
					cur_pin = ''
					cur_user = ''
					cur_selection = ''
					v.display('BYE!')
					sleep(0.5)
					mk.set_message(GREETING)
					continue
				cur_selection += chr(key + ord('0'))
				mk.set_message('SELECT: '+cur_selection)
			elif len(cur_selection) == 1:
				if key == 11:
					cur_selection = ''
					mk.set_message('SELECT: ')
					continue
				else:
					cur_selection += chr(key + ord('0'))
					#make_selection(cur_selection)
					# XXX this should move somewhere else:
					if cur_selection == '55':
						v.display('GOT DOOR?')
						os.system('su - "%s" -c "dispense door"'%username)
					elif cur_selection[1] == '8':
						v.display('GOT COKE?')
						os.system('su - "%s" -c "dispense %s"'%(username, cur_selection[0]))
					else:
						v.display('HERES A '+cur_selection)
						v.vend(cur_selection)
					sleep(0.5)
					v.display('THANK YOU')
					sleep(0.5)
					cur_selection = ''
					logout_timeout = time() + 10
